////////////////////////////////////////////////////////////////////////////////
//                                                                            
// LCDHype USBGraph2 LCD Driver DLL Source Code                               
//
// Requirements: LCDHype Version 0.37 or above                                
//                                                                            
// This is a COM-Port based driver.
// Graphic routines are optimized for T6963 and KS108 displays.
//
// This driver was mainly made to test data transmission performance of
// microcontrollers with USB connection. But you can use it with a real
// serial connection too ;)
//
// If you have a USB connection, baudrate in config.ini is not important.
// Use 9600,19200,.... or whatever you want. Data transmission speed is
// always USB transmission speed.
//
// Used devices until now:
// PIC18F2550 and T6963
// ATMega128 and FTDI FT245BM and T6963, smaller AVR/ATMega with 512 Byte RAM should work too
// PIC18F4550 and KS108 (very slow because using a SetPixel() routine in PIC)
//
//                                                                            
// hk@holger-klabunde.de                                                      
// http://www.holger-klabunde.de                                              
//
// Todo:
// COM-Port routines are blocking ! Maybe change to Overlapped IO or as
// Manatarms said "start a new thread" is faster.
// Have to make a ConfigDialog without MFC :(.
// MAKE a good documentation :((
// Commands for setting free pins of the microcontroller to switch something on or off !
// Read data from the microcontroller ! IR-Remote signals, temperature or something like that.
// 
//
// 16.02.06 Started support for KS108 Displays
//          Change in config.ini
//          LCD_Controller=6963		for T6963
//          LCD_Controller=108		for KS108
//
// 09.02.06 New Command for setting up to 8 output pins of your microcontroller
//          Does not work correctly ! Have to ask Virustrinity about this.
//
// 31.01.06 LCD_FontPitch can be changed in config.ini to 6 or 8
//
//
// 25.01.06 Optimized LCD_SendToGfxMemory()
//          Only flush gfx buffer if part of command or parameter was send.
//          If not keep command,parameter and data in buffer for later use
//          and speed up USB data transmission.
//
// 24.01.06 Added a routine to send bitmaps if drawing is faster
//          Only change contrast or backlight if values change
//
// 22.01.06 New faster LCD_SendToGfxMemory(). Thanks to Manatarms !
//          Removed very slow SET_PIXEL.
//
// 19.01.06 Fixed bug if lcdhype sets only one pixel in LCD_SendToGfxMemory()
//
// 15.01.06 Graphics can be drawn at every pixel position now
//
// 14.01.06 Made values of textbuffersize and gfxbuffersize variabel
//          Can be changed in config.ini in controller directory now.
//          Use gfxbuffersize of 128 for PIC18F2550
//          Use gfxbuffersize of 256 for FTDI FT245BM
//          textbuffersize from 16 to 32 is good for a 240x64 display
//
// 13.01.06 First Release
//
// This is FREE SOFTWARE
// Use at your own risk !                                                   
////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <windows.h>
#include <stdio.h>
#include <direct.h>

#include "lcdllc.h"
#include "serial.h"

//Supported Controller Types
#define T6963	6963
#define KS108	108

//###########################################################################
//My own command set which you have to program into YOUR microcontroller.
#define CLEAR_SCREEN	0xB0 //Clear text and graphics screen
#define SET_CURSOR		0xB1 //Set Text cursor first col=1, first row=1
#define CURSOR_HOME		0xB2 //Set text cursor to position col=1,row=1
#define SET_CONTRAST	0xB3 //0..255, 0=off
#define SET_LIGHT		0xB4 //0..255, 0=off
#define SET_CGCHAR		0xB5 //0..31 YOU HAVE 32 custom character chars with T6963 !
#define SET_PIXEL		0xB6 //Set single pixel, not used anymore
#define SET_GFX			0xB7 //Set pixel array, 6 or 8 pixels in a byte at once.
#define CLEAR_TEXT		0xB8 //Clear text screen only, not used
#define CLEAR_GFX		0xB9 //Clear graphics screen only, not used
#define SET_GFXBUFFER	0xBA //Set multiple pixel arrays of ONE row, 6 or 8 pixel per byte
#define SET_BITMAP		0xBB //Transmit a bitmap up to 255 bytes per row and up to 255 rows

#define SET_OUTPUTS		0xBE //Set or reset up to 8 pins of your microcontroller
                             //First parameter is number of the port
                             //Second parameter 8 bits of data for the port

#define LAST_COMMAND	0xBF //end of commands marker for the microcontroller

#define GFXBUFFER_FILLING		0 //hold commands,parameters and data in buffer until buffer overflow
#define GFXBUFFER_FORCE_FLUSHING	1 //if commands or parameters were send, so flush data also
                                      //to avoid mixing gfx and text data

int gfx_status=GFXBUFFER_FILLING;
int gfx_force_flush=0;			//force flushing gfxbuffer at end of LCD_SendToGfxMemory() if set to 1

#define DEBUG_OUTPUT
//#undef DEBUG_OUTPUT		//remove code for writing debug output to a file

            //LCD Driver DLL Identifier String
const char* ID            = "Holgi's USB-Graph2 Display Driver V1.09b";

            //maximum Pixels of graphics mode, 76800 = 320x240
const int   MAX_RESPIXELS = 76800;
      
//Custom Char Data Structure
struct TCharData 
{ BYTE position;   //Position of custom char in ram
  BYTE data[8][8]; //custom char 8x8 pixel data matrix
};

//LCD Driver Information Structure
struct TDLLInfo
{
  char IDArray[256];          //holds identifier of DLL
  bool SupportGfxLCD;         //does this driver support graphical LCDs?
  bool SupportTxtLCD;         //does this driver support text
  bool SupportLightSlider;    //does this driver support the light control slider
  bool SupportContrastSlider; //does this driver support the contrast slider
  bool SupportOutports;		  //does this driver support outports for controlling external circuits
  BYTE CCharWidth;            //custom char width in pixels
  BYTE CCharHeight;           //custom char height in pixels
  BYTE FontPitch;             //fontpitch of LCD in pixels
};

//Structure for Symbol LCDs
struct TSymbolInfo
{ 
  bool SymbolData[20]; //is symbol x displayed or not?
};

//global storage for COM port actions
HANDLE porth=0;		//Handle to COM Port
int baudrate=9600;	//not important for USB
char line[256];		//only error messages 

//global storage for LCD propertys
bool  LCD_LightOn     = true;  //is lcd light on?
char LCD_Port[32]     = "COM1"; //port where lcd is connected to
DWORD LCD_Exectime    = 10;    //execution time of lcd in textmode
DWORD LCD_ExectimeGfx = 10;    //execution time of lcd in graphics mode
DWORD LCD_Width       = 240;   //width of LCDs graphics mode in pixels
DWORD LCD_Height      = 64;   //height of LCDs graphics mode in pixels 
DWORD LCD_Cols        = 40;    //number of chars per line in text mode
DWORD LCD_Rows        = 8;    //number of LCD lines in text mode
LONG  LCD_FontPitch    = 6;    //fontpitch of LCD in pixels   
bool  LCD_IsReady     = true;  //is lcd ready for next operation ?
BYTE  LCD_ContrastValue = 0;    
BYTE  LCD_LightValue    = 0;
DWORD LCD_Outports      = 0;
int LCD_UseOutPorts     = 0; //send commands for microcontroller ports on=1, 0=off. change in config.ini
int LCD_Controller      = T6963; //Default T6963. Change to 6963 or 108 in config.ini

char ininame[_MAX_PATH]=""; //Die INI-Datei
//char drive[_MAX_DRIVE]=""; //Hilfsarrays
char dir[_MAX_DIR]="";
//char name[_MAX_FNAME]="";
//char ext[_MAX_EXT]="";

#ifdef DEBUG_OUTPUT
//For debugging :(
FILE *log=NULL;
int makelog=0;
char logname[_MAX_PATH];
char logbuff[1024];
#endif// DEBUG_OUTPUT

//####################################################################################################
int ReadConfigFile()
//####################################################################################################
{

 baudrate=GetPrivateProfileInt("USBGraph2","BaudRate",19200,ininame);

 textbuffersize=GetPrivateProfileInt("USBGraph2","TextBufferSize",16,ininame);
 if(textbuffersize<1 || textbuffersize>TEXT_BUFFER_SIZE) textbuffersize=16;

 gfxbuffersize=GetPrivateProfileInt("USBGraph2","GfxBufferSize",64,ininame);
 if(gfxbuffersize<1 || gfxbuffersize>GFX_BUFFER_SIZE) gfxbuffersize=64;

 gfx_force_flush=GetPrivateProfileInt("USBGraph2","GfxForceFlush",0,ininame);
 if(gfx_force_flush!=0 && gfx_force_flush!=1) gfx_force_flush=0;

 LCD_FontPitch=GetPrivateProfileInt("USBGraph2","LCD_FontPitch",6,ininame);
 if(LCD_FontPitch!=6 && LCD_FontPitch!=8) LCD_FontPitch=6;

 LCD_UseOutPorts=GetPrivateProfileInt("USBGraph2","LCD_UseOutPorts",0,ininame);
 if(LCD_UseOutPorts!=0 && LCD_UseOutPorts!=1) LCD_UseOutPorts=0;

 LCD_Controller=GetPrivateProfileInt("USBGraph2","LCD_Controller",T6963,ininame);
 if(LCD_Controller!=T6963 && LCD_Controller!=KS108) LCD_Controller=T6963; //default

 if(LCD_Controller==KS108) LCD_FontPitch=8; //for KS108 force to 8 bits

#ifdef DEBUG_OUTPUT
 makelog=GetPrivateProfileInt("USBGraph2","LogFile",0,ininame);
 if(makelog!=0 && makelog!=1) makelog=0;
#endif// DEBUG_OUTPUT

 return 0;
}


//####################################################################################################
int WriteConfigFile()
//####################################################################################################
{
 _itoa(baudrate,line,10);
 WritePrivateProfileString("USBGraph2","BaudRate",line,ininame);
 _itoa(textbuffersize,line,10);
 WritePrivateProfileString("USBGraph2","TextBufferSize",line,ininame);
 _itoa(gfxbuffersize,line,10);
 WritePrivateProfileString("USBGraph2","GfxBufferSize",line,ininame);
 _itoa(gfx_force_flush,line,10);
 WritePrivateProfileString("USBGraph2","GfxForceFlush",line,ininame);
 _itoa(LCD_FontPitch,line,10);
 WritePrivateProfileString("USBGraph2","LCD_FontPitch",line,ininame);
 _itoa(LCD_UseOutPorts,line,10);
 WritePrivateProfileString("USBGraph2","LCD_UseOutPorts",line,ininame);
 _itoa(LCD_Controller,line,10);
 WritePrivateProfileString("USBGraph2","LCD_Controller",line,ininame);

#ifdef DEBUG_OUTPUT
 _itoa(makelog,line,10);
 WritePrivateProfileString("USBGraph2","LogFile",line,ininame);
#endif// DEBUG_OUTPUT

 return 0;
}


//####################################################################################################
//function used to send data to lpt port based on lcd controller
void Send(ULONG Port, 
          BYTE High,
          BYTE Low,
          BYTE Value,
          DWORD Exectime)
//####################################################################################################
{ 
#ifdef DEBUG_OUTPUT
     if(makelog && log)
	 {
      sprintf(logbuff,"\nSend()");
      fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
	 }
#endif// DEBUG_OUTPUT

//  DWORD i;

//  DlPortWritePortUchar(LCD_Port+2,High+((BYTE)(!LCD_LightOn))*Light_Byte);
//  DlPortWritePortUchar(LCD_Port,Value);
//  for (i=0;i<=Exectime;i++) { DlPortReadPortUchar(0x220); };
//  DlPortWritePortUchar(LCD_Port+2,Low+((BYTE)(!LCD_LightOn))*Light_Byte);
//  for (i=0;i<=Exectime;i++) { DlPortReadPortUchar(0x220); };
}

//####################################################################################################
//function used to create a lcd driver config dialog
//dont export this function if you dont want to create config dialog
void _stdcall LCD_ConfigDialog()
//####################################################################################################
{
 sprintf(line,"Baudrate %d\n"
	          "TextBufferSize %d\n"
			  "GfxBufferSize %d\n"
			  ,baudrate
			  ,textbuffersize
			  ,gfxbuffersize
			 );
 MessageBox(NULL,line,"Settings",MB_OK);

// WriteConfigFile();
};

//function used to send a byte to lcd controller
//####################################################################################################
void _stdcall LCD_SendToController(BYTE Value)
//####################################################################################################
{ };

//####################################################################################################
//function used to send a byte to lcd memory
//or better write a byte on the display
void _stdcall LCD_SendToMemory(BYTE Value)
//####################################################################################################
{
 LCD_IsReady=false;
 SerialPutToTextBuffer(&porth,Value);

#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
      if(Value<0x20) 
	  {
	   sprintf(logbuff,"0x%02X",(int)Value);
       fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
	  }
      else fwrite(&Value,sizeof(BYTE),1,log); //write to debug.log

   }
#endif// DEBUG_OUTPUT

 LCD_IsReady=true;
};

//####################################################################################################
//function used to send gfx data to graphical lcd
//
//      This function is for displays up to 256x256 pixels only
//      It is optimized for T6963 displays !
//      Draws rows from up to down, and coloumns from left to right.
//
// It's possible that lcdhype sends only ONE pixel to this routine !
// It's also possible that lcdhype sends the whole display as a bitmap.
//
void _stdcall LCD_SendToGfxMemory(BYTE Pixels[MAX_RESPIXELS], //pixel color data
                                  WORD x1,                    //refresh area x1,y1,x2,y2
                                  WORD y1, 
                                  WORD x2,
                                  WORD y2,
                                  bool inverted               //output of gfx data inverted?
															//this parameter is not used by this function
                                 )
//####################################################################################################
{
 unsigned char by,i,mask,tmp,x,y,xstart,xend,ystart,yend,rest,numbytes,pages;

 bool send_bitmap;
 unsigned char bmpheight;


 LCD_IsReady=false;

 SerialFlushTextBuffer(&porth); //first send and clean text buffer to update display

#ifdef DEBUG_OUTPUT
 if(makelog && log)
 {
  sprintf(logbuff,"\nLCD_SendToGfxMemory() "
                  "x1=%03d x2=%03d y1=%03d y2=%03d\n"
				  ,(int)x1,(int)x2
				  ,(int)y1,(int)y2
	      );
  fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
 }
#endif// DEBUG_OUTPUT

//+++++++++++++++++++++++++++++++++++++++++++++++++++
// Begin KS108 routines
// Always send a whole screen for KS108
// You have to use %Param.RenderMode(Simple or Full)
// in your script header.
//+++++++++++++++++++++++++++++++++++++++++++++++++++
 if(LCD_Controller==KS108)
 {
  //KS108 displays have more than one controller.
  //Each controller handles 64x64 Pixels.
  //Pixels can be send as a page of 64x8 pixels.
  //So each controller has 8 pages.
  //Pixels in a byte are oriented vertically !
  //Bit D0 is the upper pixel in the byte.

  pages = (unsigned char)(LCD_Height / 8);
  xstart=0; xend=(unsigned char)LCD_Width;
  ystart=0; yend=(unsigned char)LCD_Height;
  numbytes=(unsigned char)(LCD_Width/8);
  bmpheight=(unsigned char)LCD_Height;

  SerialPutToGfxBuffer(&porth,SET_BITMAP); //command for microcontroller to receive a bitmap
  SerialPutToGfxBuffer(&porth,xstart); 
  SerialPutToGfxBuffer(&porth,ystart);
  SerialPutToGfxBuffer(&porth,numbytes);
  SerialPutToGfxBuffer(&porth,bmpheight);

  if(gfxbuffercount<5)                     //part of command or parameter send ?
	 gfx_status=GFXBUFFER_FORCE_FLUSHING;  //than force flushing data too !

#ifdef DEBUG_OUTPUT
     if(makelog && log)
	 {
      sprintf(logbuff,"SetBitmap() x=%03d y=%03d width=%03d bytes height=%03d pixels\n"
		      ,(int)xstart,(int)ystart,(int)numbytes,(int)bmpheight);
      fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
	 }
#endif// DEBUG_OUTPUT

  for(y=ystart; y<yend; y+=8) //row loop
  {
   for(x=xstart; x<xend; x++) //col loop
   {
    mask = 0x01; //start with LSB
	by=0;

    for(i=0; i<8; i++) //put multiple pixels in one byte loop
	{
     tmp=Pixels[(unsigned int)(y + i) * LCD_Width + (unsigned int)(x)];
     if(tmp>0) by|=mask;
	 mask<<=1;
	}

    SerialPutToGfxBuffer(&porth,by);
    if(gfxbuffercount==1)                   //part of command or parameter send ?
	  gfx_status=GFXBUFFER_FORCE_FLUSHING;  //than force flushing data too !

   }//for(x=xstart; x<xend; x++) //col loop
  }//for(y=ystart, y<yend; y+=8) //col loop
 
 }//if(LCD_Controller==KS108)
//+++++++++++++++++++++++++++++++++++++++++++++++++++
// End KS108 routines
//+++++++++++++++++++++++++++++++++++++++++++++++++++

//+++++++++++++++++++++++++++++++++++++++++++++++++++
// Begin T6963 routines
//+++++++++++++++++++++++++++++++++++++++++++++++++++
 if(LCD_Controller==T6963)
 {
 //make sure drawing from up to down
 if(y2<y1) { ystart=(unsigned char)y2; yend=(unsigned char)y1; }
 else  { ystart=(unsigned char)y1; yend=(unsigned char)y2; }

 //make sure drawing from left to right
 if(x2<x1) { xstart=(unsigned char)x2; xend=(unsigned char)x1; }
 else  { xstart=(unsigned char)x1; xend=(unsigned char)x2; }

 //######################################################################
 //Manatarms cool optimization !
 //Force start drawing from a LCD_FontPitch boundary 
 //and end at a LCD_FontPitch-1 position.
 //We can do this because Pixels[] always contains the pixels
 //of a COMPLETE display.
 //######################################################################
 while((xstart % LCD_FontPitch) != 0) xstart--;
 while((xend % LCD_FontPitch) != (LCD_FontPitch-1)) xend++;

 send_bitmap=false;

 numbytes=(unsigned char)((xend-xstart+1)/LCD_FontPitch); //number of remaining bytes per row
 bmpheight= yend-ystart+1;

 //if data has more than one row send it as a bitmap
 //each row of the bitmap can contain up to 255 bytes
 if(bmpheight>1 && numbytes<256) send_bitmap=true;

 if(send_bitmap==true)
 {
  SerialPutToGfxBuffer(&porth,SET_BITMAP); //command for microcontroller to receive a bitmap
  SerialPutToGfxBuffer(&porth,xstart);
  SerialPutToGfxBuffer(&porth,ystart);
  SerialPutToGfxBuffer(&porth,numbytes);
  SerialPutToGfxBuffer(&porth,bmpheight);

  if(gfxbuffercount<5)                     //part of command or parameter send ?
	 gfx_status=GFXBUFFER_FORCE_FLUSHING;  //than force flushing data too !

#ifdef DEBUG_OUTPUT
     if(makelog && log)
	 {
      sprintf(logbuff,"SetBitmap() x=%03d y=%03d width=%03d bytes height=%03d pixels\n"
		      ,(int)xstart,(int)ystart,(int)numbytes,(int)bmpheight);
      fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
	 }
#endif// DEBUG_OUTPUT
 }

 for(y=ystart; y<=yend; y++) //row loop
 {
  x=xstart;         //starting x position for the row

  rest=xend-xstart+1; //number of pixels to plot on the row

  //send a bitmap
  if(send_bitmap==true)
  {
	do
	{
     mask=(1<<(LCD_FontPitch-1));
     by=0;

     for(i=0; i<LCD_FontPitch; i++) //put multiple pixels in one byte loop
	 {
      tmp=Pixels[(unsigned int)y * LCD_Width + (unsigned int)(x + i)];
      if(tmp>0) by|=mask;
	  mask>>=1;
	 }

     SerialPutToGfxBuffer(&porth,by);
     if(gfxbuffercount==1)                   //part of command or parameter send ?
	   gfx_status=GFXBUFFER_FORCE_FLUSHING;  //than force flushing data too !

	 rest-=(unsigned char)LCD_FontPitch;
	 x+=(unsigned char)LCD_FontPitch;
	}while(rest>=LCD_FontPitch && x<=xend);
   }

 //only one byte on the row available 
   if(numbytes==1 && send_bitmap==false)
   {
    mask=(1<<(LCD_FontPitch-1));
	by=0;

    for(i=0; i<LCD_FontPitch; i++) //put multiple pixels in one byte loop
	{
     tmp=Pixels[(unsigned int)y * LCD_Width + (unsigned int)(x + i)];
     if(tmp>0) by|=mask;
	 mask>>=1;
	}

    SerialPutToGfxBuffer(&porth,SET_GFX); //command for microcontroller to receive only one byte with pixel data
    SerialPutToGfxBuffer(&porth,x);
    SerialPutToGfxBuffer(&porth,y);
    SerialPutToGfxBuffer(&porth,by);
    if(gfxbuffercount<4)                     //part of command or parameter send ?
	   gfx_status=GFXBUFFER_FORCE_FLUSHING;  //than force flushing data too !

#ifdef DEBUG_OUTPUT
     if(makelog && log)
	 {
      sprintf(logbuff,"SetGfxSingle() x=%03d y=%03d by=0x%02X\n",(int)x,(int)y,(int)by);
      fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
	 }
#endif// DEBUG_OUTPUT

	rest-=(unsigned char)LCD_FontPitch;
	x+=(unsigned char)LCD_FontPitch;
   }
	  

 //multiple bytes available for ONE row 
   if(rest>LCD_FontPitch && numbytes>1 && send_bitmap==false)
   {
    SerialPutToGfxBuffer(&porth,SET_GFXBUFFER); //command for the microcontroller to receive a LOT of bytes
    SerialPutToGfxBuffer(&porth,x);             //x startpositon of the row
    SerialPutToGfxBuffer(&porth,y);             //y the row to draw data 
    SerialPutToGfxBuffer(&porth,numbytes);      //number of bytes for graphics data to receive for the microcontroller

    if(gfxbuffercount<4)                     //part of command or parameter send ?
	   gfx_status=GFXBUFFER_FORCE_FLUSHING;  //than force flushing data too !
    
#ifdef DEBUG_OUTPUT
     if(makelog && log)
	 {
      sprintf(logbuff,"SetGfxMulti() x=%03d y=%03d bytes=%03d\n",(int)x,(int)y,(int)numbytes);
      fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
	 }
#endif// DEBUG_OUTPUT

	do
	{
     mask=(1<<(LCD_FontPitch-1));
 	 by=0;

     for(i=0; i<LCD_FontPitch; i++) //put multiple pixels in one byte loop
	 {
      tmp=Pixels[(unsigned int)y * LCD_Width + (unsigned int)(x + i)];
      if(tmp>0) by|=mask;
	  mask>>=1;
	 }


     SerialPutToGfxBuffer(&porth,by);
     if(gfxbuffercount==1)                   //part of command or parameter send ?
	   gfx_status=GFXBUFFER_FORCE_FLUSHING;  //than force flushing data too !

 	 rest-=(unsigned char)LCD_FontPitch;
	 x+=(unsigned char)LCD_FontPitch;

	}while(rest>=LCD_FontPitch && x<=xend);
   }

 } //end row loop
 }//if(LCD_Controller==T6963)
//+++++++++++++++++++++++++++++++++++++++++++++++++++
// End T6963 routines
//+++++++++++++++++++++++++++++++++++++++++++++++++++

 if(gfx_status==GFXBUFFER_FORCE_FLUSHING || gfx_force_flush!=0)
 {
  //if comands or parameters were send to this point, it is necessary to
  //flush the gfxbuffer. EVEN if only ONE byte of data is in the buffer !
  //this has to be done to avoid mixing text and gfx buffers.
  SerialFlushGfxBuffer(&porth); //send gfx buffer
  gfx_status=GFXBUFFER_FILLING;
 }

 LCD_IsReady=true;
};

//####################################################################################################
//function to set the output address in text mode
//set the cursor for next char position to x (coloumn),y (row)
//x and y are zero based positions !
void _stdcall LCD_SetOutputAddress(DWORD x, DWORD y)
//####################################################################################################
{
 LCD_IsReady=false;

 SerialPutToTextBuffer(&porth,SET_CURSOR);
 SerialPutToTextBuffer(&porth,(unsigned char)x+1); //my display has 1 based coloumns
 SerialPutToTextBuffer(&porth,(unsigned char)y+1); //my display has 1 based rows

#ifdef DEBUG_OUTPUT
 if(makelog && log)
 {
  sprintf(logbuff,"\nSetTextCursor() "
                  "x=%03d y=%03d\n"
                  ,(int)x,(int)y
	      );
  fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
 }
#endif// DEBUG_OUTPUT

 LCD_IsReady=true;

};

//####################################################################################################
//function to write a custom char to ram
//each custom char comes in an 8x8 PIXEL matrix !
//first byte is the upper row of the char
//MSB is left position.
void _stdcall LCD_SetCGRAMChar(TCharData cdata)
//####################################################################################################
{
 unsigned char by,x,y,mask;

 LCD_IsReady=false;

 SerialPutToTextBuffer(&porth,SET_CGCHAR);
 SerialPutToTextBuffer(&porth,(unsigned char)cdata.position);

#ifdef DEBUG_OUTPUT
     if(makelog && log)
	 {
      sprintf(logbuff,"\nSetCGRAMChar() pos=%03d ",(int)cdata.position);
      fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
	 }
#endif// DEBUG_OUTPUT

 for(x=0; x<8; x++) // x = rows of the custom char
 {

  by=0;
 
  mask=(1<<(LCD_FontPitch-1)); //MSB first

  for(y=0; y<LCD_FontPitch; y++) // y = col of the custom char
  {
   if(cdata.data[y][x] > 0) by|=mask;

   mask>>=1;
  }

  SerialPutToTextBuffer(&porth,by);

#ifdef DEBUG_OUTPUT
  if(makelog && log)
	 {
      sprintf(logbuff,"0x%02X,",(int)by);
      fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
	 }
#endif// DEBUG_OUTPUT

 }

#ifdef DEBUG_OUTPUT
  if(makelog && log)
	 {
      sprintf(logbuff,"\n");
      fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
	 }
#endif// DEBUG_OUTPUT

  LCD_IsReady=true;
};

//####################################################################################################
//function to get the position of custom char
//IN : custom char position based on HD44780 LCDs (0...)
//OUT: custom char position converted to current lcd controllers position
//
//Example : HD44780, first custom char is at position 0, return 0
//          T6963c , first custom char is at position 128, return position+128 !
//
//
BYTE _stdcall LCD_GetCGRAMChar(BYTE position)
//####################################################################################################
{ return position; }; 


//####################################################################################################
//function used to clean up driver usage, called when DLL is closed.
void _stdcall LCD_CleanUp()
//####################################################################################################
{
 LCD_IsReady=false;

 SerialFlushTextBuffer(&porth);
 SerialFlushGfxBuffer(&porth);

 //Clear Screen Command
 SerialPutc(&porth,CLEAR_SCREEN);

 //Contrast Off 
 SerialPutc(&porth,SET_CONTRAST);
 SerialPutc(&porth,0);
 LCD_ContrastValue	= 0;    

//Backlight off
 SerialPutc(&porth,SET_LIGHT);
 SerialPutc(&porth,0);
 LCD_LightValue		= 0;

//TODO: Clean all outports
// LCD_Outports = 0;


// Sleep(100);

#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
    sprintf(logbuff,"\nLCD_CleanUp()\n");
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT

 if(porth!=0 && porth!=INVALID_HANDLE_VALUE) //clean up the COM port
 {
	PurgeComm(porth,PURGE_RXABORT | PURGE_RXCLEAR);
    PurgeComm(porth,PURGE_TXABORT | PURGE_TXCLEAR);
	EscapeCommFunction(porth, CLRDTR);
	EscapeCommFunction(porth, CLRRTS);
    CloseHandle(porth);
    porth=0;
//    Sleep(1000);

#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
    sprintf(logbuff,"\nLCD_CleanUp() porth=closed\n");
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }

  if(log) { fclose(log); log=NULL; } //close the log file
#endif// DEBUG_OUTPUT
 }


 LCD_IsReady=true;
};

//####################################################################################################
//function used to initialize LCD
void _stdcall LCD_Init(void)
//####################################################################################################
{
  LCD_IsReady=false;

  _getcwd(dir,_MAX_DIR); //get current working directory
  _makepath(ininame,NULL,dir,"config","ini");

#ifdef DEBUG_OUTPUT
  _makepath(logname,NULL,dir,"debug","log");
#endif// DEBUG_OUTPUT

  ReadConfigFile();
  WriteConfigFile(); //write default config.ini in controller directory if not exists

#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
    sprintf(logbuff,"\nLCD_Init()\n");
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT

//lcdhype calls this function more than one time !
//if porth is a valid handle, than CLOSE it first,
//or WIN will report "Access denied" for the already
//open COM-Port !
 if(porth!=0) LCD_CleanUp();

 if(porth==0)
 {
#ifdef DEBUG_OUTPUT
  //IMPORTANT NOTICE:
  //Delete last debug.log if you want to make a NEW debug.log file !
  //Data is appended to the last logfile !
  if(makelog && log==NULL) log=fopen(logname,"a+t"); //open debug log file

   if(makelog && log)
   {
    sprintf(logbuff,"\nLCD_Init() porth=closed\n");
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT

  porth=SerialInit(LCD_Port, baudrate); 
  if(porth==0)
  {
//  sprintf(line,"##################### '%s' not available ! #####################",LCD_Port);
//  MessageBox(NULL,line,NULL,MB_OK);
   LCD_IsReady=true;
   return;
  }

#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
    sprintf(logbuff,"LCD_Init() porth=open\n");
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT

 }

 SerialPutc(&porth,CLEAR_SCREEN);
//    Sleep(1000);

//TODO: Clean all outports
// LCD_Outports = 0;

 gfx_status=GFXBUFFER_FILLING;

 LCD_IsReady=true;
};

//####################################################################################################
//function used to indicate that the lcd can receive data
//commonly used when interfacing serial lcds else return value true
bool _stdcall LCD_IsReadyToReceive()
//####################################################################################################
{
#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
    sprintf(logbuff,"\nLCDHype called LCD_IsReadyToReceive()\n");
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT

 return LCD_IsReady;
}; 



//####################################################################################################
//function used to set the lcd propertys, so that the lcd driver can work correctly
//commonly used when supporting more than one lcd type in one driver
void _stdcall LCD_SetIOPropertys(char* Port,                //port lcd is connected to
								 DWORD Exectime,            //execution time for text mode
								 DWORD ExectimeGfx,         //execution time for gfx mode
								 DWORD X,                   //cols in text mode
								 DWORD Y,                   //rows in text mode
								 DWORD GX,                  //width in pixels in gfx mode
								 DWORD GY,                  //height in pixels in gfx mode
								 bool  LightOn,             //is light control enabled?
								 BYTE  LightSliderValue,    //brightness value from light slider
								 bool  ContrastOn,          //is contrast control enabled?
								 BYTE  ContrastSliderValue, //contrast value from contrast slider
								 DWORD Outports,            //32bit = 32 outports, bit=0=off, bit=1=on 
								 bool  UnderlineMode,       //is underline mode in text mode enabled?
								 bool  UnderlineOutput      //display underlined text?
								)
//####################################################################################################
{ 
 //This function is called periodically by lcdhype ?

 strcpy(LCD_Port,Port);
 LCD_Exectime		= Exectime;
 LCD_ExectimeGfx	= ExectimeGfx;
 LCD_Cols			= X;
 LCD_Rows			= Y;
 LCD_Width			= GX;
 LCD_Height			= GY;
 LCD_LightOn		= LightOn;

#ifdef DEBUG_OUTPUT
  if(makelog && log)
   {
    sprintf(logbuff,"\nLCDHype called LCD_SetIOPropertys()\n");
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT

 if(LCD_LightOn && LCD_LightValue != LightSliderValue)
 {
  LCD_LightValue		= LightSliderValue;
  SerialPutToTextBuffer(&porth,SET_LIGHT);
  SerialPutToTextBuffer(&porth,LCD_LightValue);   //Backlight on

#ifdef DEBUG_OUTPUT
  if(makelog && log)
   {
    sprintf(logbuff,"\nLCD_SetIOPropertys() Light=%d\n",LCD_LightValue);
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT
 }

 if(LCD_LightOn==false)
 {
  LCD_LightValue		= 0;
  SerialPutToTextBuffer(&porth,SET_LIGHT);
  SerialPutToTextBuffer(&porth,LCD_LightValue);   //Backlight on

#ifdef DEBUG_OUTPUT
  if(makelog && log)
   {
    sprintf(logbuff,"\nLCD_SetIOPropertys() Light=%d\n",LCD_LightValue);
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT
 }

 if( LCD_ContrastValue	!= ContrastSliderValue )
 {
  LCD_ContrastValue	= ContrastSliderValue;    
  SerialPutToTextBuffer(&porth,SET_CONTRAST);
  SerialPutToTextBuffer(&porth,LCD_ContrastValue);

#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
    sprintf(logbuff,"\nLCD_SetIOPropertys() SetContrast=%d\n",LCD_ContrastValue);
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT
 }

//############### DON'T USE THIS ! BUGGY ######################
//LCDHype can send up to 4 bytes (32 pins) to switch pins of a microcontroller on or off
//IMPORTANT: You have to ignore values not used by your microcontroller in your firmware ! 
 if( LCD_Outports != Outports && LCD_UseOutPorts>0 )
 {
  //First byte for outports if value has changed
  if((LCD_Outports&0x000000FF) != (Outports&0x000000FF))
  {

   SerialPutToTextBuffer(&porth,SET_OUTPUTS);
   SerialPutToTextBuffer(&porth,0); //number of port. 
   SerialPutToTextBuffer(&porth,(unsigned char)(LCD_Outports&0xFF)); //pin values. 

#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
    sprintf(logbuff,"\nLCD_SetIOPropertys() Port0=0x%02X\n",(int)(LCD_Outports&0xFF));
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT
  }//First byte for outports

  //2nd byte for outports if value has changed
  if((LCD_Outports&0x0000FF00) != (Outports&0x0000FF00))
  {

   SerialPutToTextBuffer(&porth,SET_OUTPUTS);
   SerialPutToTextBuffer(&porth,1); //number of port. 
   SerialPutToTextBuffer(&porth,(unsigned char)((LCD_Outports>>8)&0xFF)); //pin values. 

#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
    sprintf(logbuff,"\nLCD_SetIOPropertys() Port1=0x%02X\n",(int)((LCD_Outports>>8)&0xFF));
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT
  }//2nd byte for outports

  //3rd byte for outports if value has changed
  if((LCD_Outports&0x00FF0000) != (Outports&0x00FF0000))
  {

   SerialPutToTextBuffer(&porth,SET_OUTPUTS);
   SerialPutToTextBuffer(&porth,2); //number of port. 
   SerialPutToTextBuffer(&porth,(unsigned char)((LCD_Outports>>16)&0xFF)); //pin values. 

#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
    sprintf(logbuff,"\nLCD_SetIOPropertys() Port2=0x%02X\n",(int)((LCD_Outports>>16)&0xFF));
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT
  }//3rd byte for outports

  //4th byte for outports if value has changed
  if((LCD_Outports&0xFF000000) != (Outports&0xFF000000))
  {

   SerialPutToTextBuffer(&porth,SET_OUTPUTS);
   SerialPutToTextBuffer(&porth,3); //number of port. 
   SerialPutToTextBuffer(&porth,(unsigned char)((LCD_Outports>>24)&0xFF)); //pin values. 

#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
    sprintf(logbuff,"\nLCD_SetIOPropertys() Port3=0x%02X\n",(int)((LCD_Outports>>24)&0xFF));
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT
  }//4th byte for outports

  LCD_Outports = Outports;
  SerialFlushTextBuffer(&porth); //It's better to do this here
 }

};

//function used to interface symbols of a symbol lcd
void _stdcall LCD_Customize(TSymbolInfo SymData)
{ };

//####################################################################################################
//function used to return information about the dll to the calling process
TDLLInfo _stdcall DLL_GetInfo()
//####################################################################################################
{ 
#ifdef DEBUG_OUTPUT
   if(makelog && log)
   {
    sprintf(logbuff,"\nLCDHype called DLL_GetInfo()\n");
    fwrite(logbuff,strlen(logbuff),1,log); //write to debug.log
   }
#endif// DEBUG_OUTPUT

  TDLLInfo result;  
//  result.CCharWidth  = 6;       
  result.CCharWidth  = (unsigned char)LCD_FontPitch;       
  result.CCharHeight = 8;
//  result.FontPitch = 6;
  result.FontPitch = (unsigned char)LCD_FontPitch;
  result.SupportGfxLCD = true;
  result.SupportTxtLCD = true;
  result.SupportLightSlider = true;
  result.SupportContrastSlider = true;
  result.SupportOutports = true;

  ZeroMemory(&result.IDArray,256);
  CopyMemory(&result.IDArray[0],&ID[0],strlen(ID));

  return result;
};

